const pipe = require('./pipe')
const compose = require('./compose')
const tap = require('./tap')
const all = require('./all')
const assign = require('./assign')
const tryCatch = require('./tryCatch')
const switchCase = require('./switchCase')
const map = require('./map')
const filter = require('./filter')
const reduce = require('./reduce')
const transform = require('./transform')
const flatMap = require('./flatMap')
const forEach = require('./forEach')
const some = require('./some')
const every = require('./every')
const and = require('./and')
const or = require('./or')
const not = require('./not')
const eq = require('./eq')
const gt = require('./gt')
const lt = require('./lt')
const gte = require('./gte')
const lte = require('./lte')
const get = require('./get')
const set = require('./set')
const pick = require('./pick')
const omit = require('./omit')
const thunkify = require('./thunkify')
const always = require('./always')
const curry = require('./curry')
const __ = require('./__')

/**
 * design principles
 *
 * rubico is a module, not a grab bag
 * functional code should not care about async
 * exported methods are time and space optimal
 * memory used by exported methods is properly garbage collected
 * no special types; use built-in types
 * avoid variadic functions; use lists
 * avoid anonymous function creation; use names and factory functions
 * avoid creating functions inside functions
 */

const rubico = {
  pipe, compose,

  tap, forEach,

  switchCase,

  tryCatch,

  all, assign, get, set, pick, omit,

  map, filter, flatMap, reduce, transform,

  and, or, not, some, every,

  eq, gt, lt, gte, lte,

  thunkify, always, curry, __,
}

module.exports = rubico
